from typing import List

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.preprocessing import OrdinalEncoder

show_plots = False

def ordinal_encode(df, categories, target):
    """
    Applies ordinal encoding to a specified categorical column in a DataFrame.

    Parameters:
    df (pd.DataFrame): The input DataFrame.
    categories (list): A list containing the ordered categories for encoding.
    target (str): The column name to be encoded.

    Raises:
    TypeError: If categories are not provided or do not contain exactly two elements.

    Modifies:
    The function modifies the input DataFrame by replacing the target column with its encoded values.
    """
    if categories is None or len(categories) != 2:
        raise TypeError("Categories must be provided")
    df[target] = OrdinalEncoder(categories = [categories]).fit_transform(df[[target]])

def heat_map(df, model_name=None):
    """
    Generates a heatmap of the correlation matrix for numerical features in the DataFrame.

    Parameters:
    df (pd.DataFrame): The input DataFrame.

    Modifies:
    Displays a heatmap visualization of the feature correlations.
    """
    # Drop all NaN
    df.dropna(axis='columns')
    ndf = df[[col for col in df.columns if df[col].nunique() > 1 and pd.api.types.is_numeric_dtype(df[col])]]
    corr = ndf.corr()
    plt.figure(figsize=(15, 12))
    sns.heatmap(corr)
    if model_name:
        save_plot(model_name + " - Heat Map")
    if show_plots:
        plt.show()

def plot_counts(model_name=None):
    if model_name:
        save_plot(model_name + " - Counts")
    if show_plots:
        plt.show()

def plot_xy(df, x, y, model_name=None):
    """
    Creates a scatter plot for two numerical columns.

    Parameters:
    df (pd.DataFrame): The input DataFrame.
    x (str): The column name for the x-axis.
    y (str): The column name for the y-axis.

    Modifies:
    Displays a scatter plot of the two selected features.
    """
    plt.scatter(df[x], df[y])
    plt.xlabel(x)
    plt.ylabel(y)
    if model_name:
        save_plot(model_name + " - XY")
    if show_plots:
        plt.show()

def plot_features(features, info_text: str = None, model_name=None):
    """
    Parameters:
    columns (list): The list of feature names used in the model.

    Modifies:
    Displays a horizontal bar chart representing feature importance.
    """
    plt.figure(figsize=(10, 10))

    # Add labels to bars
    for index, value in enumerate(features['Importance']):
        plt.text(value + 0.005, index, f'{value:.4f}', fontsize=10, verticalalignment='center')

    plt.barh(features.index, features.Importance)
    plt.xlabel('Importance')
    plt.title('Feature Importance')
    plt.gca().invert_yaxis()

    # Fix text cutoff issue
    plt.subplots_adjust(left=0.3)  # ⬅ Increase left margin for feature names
    plt.tight_layout()

    if info_text:
        x_max = plt.xlim()[1]
        y_max = plt.ylim()[0]
        plt.text(x_max, y_max, info_text, verticalalignment='bottom', horizontalalignment='right')
    if model_name:
        save_plot(model_name + " - Features")
    if show_plots:
        plt.show()

def normalize(df, exclude, numerical_scaler, label_scaler):

    scale_targets = df.select_dtypes(include=np.number).drop(columns=exclude).columns
    df[scale_targets] = numerical_scaler.fit_transform(df[scale_targets])

    labels = df.select_dtypes(include=object, exclude=np.number).columns
    for label in labels:
        df[label] = label_scaler.fit_transform(df[label])

def plot_confusion_matrix(confusion_matrix: List[List[int]], accuracy: float, model_name=None) -> None:
    if len(confusion_matrix) != 2 or any(len(row) != 2 for row in confusion_matrix):
        raise ValueError("Confusion matrices must be 2x2")

    plt.figure(figsize=(8, 6))
    labels = [[f'True Positive\n{confusion_matrix[0][0]}', f'False Positive\n{confusion_matrix[0][1]}'],
             [f'False Negative\n{confusion_matrix[1][0]}', f'True Negative\n{confusion_matrix[1][1]}']]
    plt.title(f"Confusion matrix | Accuracy: {accuracy * 100:.2f}%")
    sns.heatmap(confusion_matrix, annot=labels, fmt='', cmap='Blues', cbar=False)
    if model_name:
        save_plot(model_name + " - Confusion Matrix")
    if show_plots:
        plt.show()

def print_high_confidence_samples(model, x: pd.DataFrame):
    # Get predicted probabilities
    predicted_probabilities = pd.DataFrame(model.predict_proba(x)[:, 1],
                                           columns=['confidence level'])  # Probability of being class 1
    # Filter samples where the model is at least 90% sure
    high_confidence_samples = predicted_probabilities[predicted_probabilities['confidence level'] > 0.9]
    print(high_confidence_samples.head())

def save_plot(name):
    plt.savefig("resulting_figures/" + name, dpi=300, bbox_inches='tight')
